; Simple LED dimmer.  File: dim_spwm.asm
;
; Global Assembler directives
	list		c=132, n=80
	list		p=12F615
	errorlevel	-302		; Disable messages for register banking
	errorlevel	-306		; Disable messages for page crossing

#include		p12f615.inc

;
; This program controls a single LED channel.  It allows the intensity of the LED
; channel to be controlled using 8-bit pulse-width modulation. PWM frequency is ~486 Hz.
; No attempt is made to correct the intensity to linearize the perceived brigtness of the
; LED.
;
; The intensity is set by an external voltage on Vin.  This voltage is referenced
; to VDD.  Full intensity occurs with Vin >= VDD.  Noise on Vin is handled
; by averaging multiple readings in an averaging buffer.  Values on Vin are read
; every 25 mSec.
;
; Internal processor state is occasionally refreshed in an attempt to deal with
; some ESD-related corruption (e.g. port direction bits).
;
; Copyright (c) 2009-2010 Dan Julio.  All rights reserved.
; Permission granted for personal, non-commercial use (be cool).
;
; Revision History
;   12/03/10 Revision 0.1  DJJ  First pass.  Based on w_pwm_an.asm with replacement
;                               of simple PWM algorithm.
;
; ------------------------------------------------------------------------
; Processor Information:
;
; Target processor: PIC 12C615
; Processor frequency: Built-in 8 MHz (500 ns instruction cycle)
;
; Memory Layout
;    Page 0: 0x0000 - 0x03FF    Main code, Interrupt handler, subroutines
;
; Input/Output Pin Assignments (PDIP/SOIC package)
;
;   Pin     Signal                       Polarity      I/O     Default State
;   ------------------------------------------------------------------------
;    1 VDD Power+                        -             Input   -
;    2 GP5 LED Output                    Active High   Output  0
;    3 GP4 Reserved                      Active High   Output  0
;    4 GP3 MCLR/VPP                      Active Low    Input   0 VPP
;    5 GP2 Vin                           -             Input   0
;    6 GP1 ICSP CLK                      Active High   Output  0 ICSPCLK
;    7 GP0 ICSP DAT                      Active High   Output  0 ICSPDAT
;    8 VSS Power-                        -             Input   -
;
;
; PIC Timer Utilization
; ---------------------
;   TIMER0 - used to generate the PWM clock
;   TIMER1 - used to generate a 25 mSec period for internal evaluations
;   TIMER2 - not used
;
;
; ------------------------------------------------------------------------
; File Layout:
;
;    1. Configuration Bits
;    2. Program Constants
;    3. Macro definitions
;    4. Main code
;    5. Main code subroutines
;
;
; ------------------------------------------------------------------------
; Set the device configuration bits
;  CONFIG
;    BOR = 11 (Brown-out Reset enabled)
;    IOSCFS = 1 (Internal Oscillator = 8 MHz)
;    CP = 1 (Code protection off) 
;    MCLRE = 0 (GP3 pin function is digital IO)
;    PWRTE = 0 (Power-up Timer enabled)
;    WDTE = 1 (WDT enabled)
;    FOSC2:0 = 100 (INTRC oscillator, GP4, GP5 are I/O)

        __config _BOR_ON & _IOSCFS_8MHZ & _CP_OFF & _MCLRE_ON & _PWRTE_ON & _WDT_ON & _INTRC_OSC_NOCLKOUT


; ------------------------------------------------------------------------
; Program Constants
;

;
; External device bits (port number bits)
EXT_LED		equ	.5
;
; Flags variable bits
OnPWMsegment			; Set during the "on" portion of the PWM period
;
; Variable register equates
TempReg		equ	0x40
Flags		equ	0x41
CurIntensity	equ	0x42	; Current average intensity
CurPWMValue	equ	0x43	; 8-bit intensity existing over one PWM period
PushPtr		equ	0x44	; Pointer to push new data into the averaging buffer
AvgHi		equ	0x45	; Accumulator when averaging analog input readings
AvgLo		equ	0x46
StateResetCount	equ	0x47	; Counter to trigger PIC state reset
;
; Input averaging buffer (maximum of 32 entries)
AvgBuf		equ	0x50 
;
; Interrupt state save registers
STATUS_Temp	equ	0x7D
PCLATH_Temp	equ	0x7E
W_Temp		equ	0x7F
;
;
; Initialization constants
STATUS_SETUP	equ	B'00011000'	; 7: IRP = 0 (page 0-1)
					; 6:5: RP1:0 = 00b (page 0)
					; 4: TO = 1 (default state)
					; 3: PD = 1 (default state)
					; 2: Z = 0
					; 1: DC = 0
					; 0: C = 0

OPTION_SETUP	equ	B'01000011'	; 7: GPPU = 0 (enable pull-ups)
					; 6: INTEDG = 1 (Int on rising edge)
					; 5: T0CS = 0 (internal TMR0 clock)
					; 4: T0SE = 0 (rising TMR0 clock)
					; 3: PSA = 0 (prescaler to TMR0)
					; 2:0: prescaler = 011b (1:16) for every 8 uSec

INTCON_SETUP	equ	B'10100000'	; 7: GIE = 1 (interrupts enabled)
					; 6: PEIE = 0 (Peripheral ints disabled)
					; 5: T0IE = 1 (TMR0 ints enabled)
					; 4: INTE = 0 (INT ints disabled)
					; 3: GPIE = 0 (Port ints disabled)
					; 2: T0IF = 0
					; 1: INTF = 0
					; 0: GPIF = 0

PIE1_SETUP	equ	B'00000000'	; 7: Unimplemented
					; 6: ADIE = 0 (A/D converter)
					; 5: CCP1IE = 0 (CCP1 Interrupt enable)
					; 4: Unimplemented
					; 3: CMIEE = 0 (Comparator Interrupt Enable bit)
					; 2: Unimplemented
					; 1: TMR2IE = 0 (TMR2 to PR2 match)
					; 0: TMR1IE = 0 (TMR1 Overflow)

GPIO_INIT	equ	B'00000000'	; No outputs enabled
GPIO_OE		equ	B'00001100'	; GP2, GP3 inputs

ANSEL_INIT	equ	B'01010100'	; 7: Unimplemented
					; 6:4: ADCS[2:0] = 101 (Fosc/16)
					; 3:0: ANS[3:0] = 0100 (GP2 analog inputs)

WPU_INIT	equ	B'00000000'	; No weak pull-ups enabled

IOC_INIT	equ	B'00000000'	; No interrupt-on-change enabled

T1CON_INIT	equ	B'00000001'	; 7: T1GINV = 0 (Timer1 Gate Invert bit)
					; 6: TMR1GE = 0 (Timer1 is on)
					; 5:4: T1CKPS1:0 = 00 (1:1 prescale)
					; 3: T1OSCEN = 0 (LP Oscillator shut off)
					; 2: T1SYNC = 0 (Sync control unused)
					; 1: TMR1CS = 0 (Internal clock)
					; 0: TMR1ON = 1 (Enable Timer)

ADCON0_INIT	equ	B'00001001'	; 7: ADFM = 0 (Left justified)
					; 6: VCFG = 0 (Voltage Reference = VDD)
					; 5: Unimplemented
					; 4:2: CHS[2:0] = 010 (Channel 02 - AN2)
					; 1: GO = 0
					; 0: ADON = 1 (AD on)

; TIMER1 configured to rollover every 25 mSec: 65536 - 50,000 (50,000 x 500 nSec)
TIMER1_INIT	equ	.65536 - .50000
;
; Number of Averaging buffer entries
;   Must be an even power-of-two number (2, 4, 8, 16 or 32)
NUM_AVG_BUF_ENTRIES	equ	.16
NUM_AVG_BUF_LOG		equ	.4		; Number of bits LOG2
AVG_BUF_PTR_MASK	equ	0x0F		; Mask bits based on number of bits
						;   NUM_AVG_BUF_LOG = 1 -> 0x01
						;   NUM_AVG_BUF_LOG = 2 -> 0x03
						;   NUM_AVG_BUF_LOG = 3 -> 0x07
						;   NUM_AVG_BUF_LOG = 4 -> 0x0F
						;   NUM_AVG_BUF_LOG = 5 -> 0x1F



; ------------------------------------------------------------------------
; Macro Definitions
;

SetExtLed		macro	Val
	if (Val == 0)
		bcf	GPIO,EXT_LED
	else
		bsf	GPIO,EXT_LED
	endif
	endm


SET_TIMER1		macro
	movlw		TIMER1_INIT & 0xFF
	movwf		TMR1L
	movlw		(TIMER1_INIT >> 8) & 0xFF
	movwf		TMR1H
	bcf		PIR1,TMR1IF
	endm


; ------------------------------------------------------------------------
; Code Block
; 
	; Reset Vector
PAGE_0	org		0x0
	goto		MAIN_VECTOR


; ------------------------------------------------------------------------
; Interrupt Code Block
	org		0x4

;
; Interrupt Service subroutine for LED control.  Called when the TMR0 times out.
;
; Implement a simple 8-bit PWM (255 on values, 1 off value).  The PWM period is split into
; two intervals.  The "on" interval where the LED is turned on and the "off" interval where
; the LED is turned off.  The length of each interval is controlled by setting timer 0
; according to the intensity.  The 8-bit timer counts 256 sub-intervals so we must offset
; the values we load into it by 1 to get the required 255 sub-intervals for the PWM engine.
;
;   On interval: 256 - Intensity (255 - Intensity + 1)
;   Off interval: Intensity + 1
; 
; The sub-intervals for intensities 1 and 255 are slightly longer than all other sub-
; intervals because of the overhead of the ISR.  This could be corrected with a more complex
; PWM implementation but is sufficient for this application.
; 
;    Sub-interval 1 and 255: 17 uSec
;    Other sub-intervals: 8 uSec each
;
; Total PWM period is ~2058 uSec for a PWM frequency of ~486 Hz (2*17 + 253*8 uSec)
; (based on a nominal PIC clock of 8 MHz).
;
; The PWM engine requires a maximum of 25 CPU cycles per interrupt, twice per PWM
; period for a worst case CPU consumption of 50 out of 4116 cycles or 1.2% CPU utilization.
;
; on entry: CurIntensity contains the latest average intensity
;           CurPWMValue may contain the intensity value being used for this PWM period
;           Flags.OnPWMsegment set during the on interval
;
; on exit: LED may be set or cleared
;          T0IF cleared
;          CurPWMValue may be updated (at the beginning of a PWM period)
;          Flags.OnPWMsegment may be flipped
;
	; Save interrupt state
	movwf		W_Temp
	swapf		STATUS,W
	clrf		STATUS
	movwf		STATUS_Temp

	; Determine which phase we're going to evaluate
	btfsc		Flags,OnPWMsegment
	goto		PWM_OFF_CHK

	; Starting "on" phase
	movf		CurIntensity,W
	movwf		CurPWMValue		; grab current intensity for use this interval
	btfsc		STATUS,Z		; Intensity = 0?
	goto		PWM_OFF_SET		;    Y: No "on" phase this interval
	SetExtLed	1			;    N: Start "on" phase
	comf		CurPWMValue,W		; Set Timer for the on interval
	addlw		.1			;   LED on for Intensity/255 timeouts
	movwf		TMR0

PWM_ON_EXIT
	bsf		Flags,OnPWMsegment
	bcf		INTCON,T0IF		; Clear the interrupt

	; Restore interrupt state
	swapf		STATUS_Temp,W
	movwf		STATUS
	swapf		W_Temp,F
	swapf		W_Temp,W
	; Return from the interrupt (and re-enable interrupts)
	retfie

PWM_OFF_CHK
	; Check if there is an "off" phase for this interval
	comf		CurPWMValue,W
	btfsc		STATUS,Z		; Intensity = 255?
	goto		PWM_OFF_EXIT		;   Y: No "off" phase for this interval

PWM_OFF_SET
	; start the "off" phase
	SetExtLed	0
	movf		CurPWMValue,W		; Set Timer for the off interval
	addlw		.1			;   LED off for (255-Intensity)/255 timeouts
	movwf		TMR0

PWM_OFF_EXIT
	bcf		Flags,OnPWMsegment
	bcf		INTCON,T0IF		; Clear the interrupt

	; Restore interrupt state
	swapf		STATUS_Temp,W
	movwf		STATUS
	swapf		W_Temp,F
	swapf		W_Temp,W
	; Return from the interrupt (and re-enable interrupts)
	retfie


; ------------------------------------------------------------------------
; Main Code Block
; 
MAIN_VECTOR
	; Perform initial power-on PIC initialization
	call		InitPic

	; Configure Timer1 for the first evaluation timeout
	SET_TIMER1

	; Configure Timer0 and enabled interrupts to start the PWM
	movlw		0xFE			; Timeout immediately to start PWM
	movwf		TMR0
	movlw		INTCON_SETUP
	movwf		INTCON

MAIN_LOOP
	; Spin waiting for timer1 to expire
	clrwdt
	btfss		PIR1,TMR1IF
	goto		MAIN_LOOP

	; Reset timer1
	SET_TIMER1

	; Trigger and read the ADC
	bsf		ADCON0,GO
ADC_LOOP
	btfsc		ADCON0,GO		; Spin until the conversion is done
	goto		ADC_LOOP

	; Push the entry into the averaging buffer and update the intensity
	call		EvalAverage

	; Check if it is time to reprogram the PIC state
	incf		StateResetCount,F
	btfss		STATUS,Z		; Rollover?
	goto		MAIN_LOOP		;   N: do nothing
	call		ReInitPic		;   Y: Reset PIC state

	goto		MAIN_LOOP




; ------------------------------------------------------------------------
; Subroutines
; 

; InitPic: Execute power-on processor initialization
;
; On entry: none
;
; On exit: PIC state initialized
;
InitPic
	; Initialize unique PIC control registers in page 0
	clrf		TMR0
	clrf		FSR
	movlw		STATUS_SETUP
	movwf		STATUS
	movlw		GPIO_INIT
	movwf		GPIO

	; Initialize unique PIC control registers in page 1
	bsf		STATUS,RP0	; Point to page 1 registers
	movlw		OPTION_SETUP
	movwf		OPTION_REG
	bcf		STATUS,RP0	; Point to page 0 registers again

	; Initialize the remaining PIC registers
	call		ReInitPic

	; Initialize variables and configure PWM state machine
	clrf		Flags
	clrf		CurIntensity
	clrf		CurPWMValue
	movlw		AvgBuf		; Set PushPtr to start of averaging buffer
	movwf		PushPtr
	clrf		StateResetCount

	; Clear the AvgBuf
	movlw		NUM_AVG_BUF_ENTRIES
	movwf		TempReg
	movlw		AvgBuf
	movwf		FSR
CLR_AVG_BUFF_LOOP
	clrf		INDF
	incf		FSR,F
	decfsz		TempReg,F
	goto		CLR_AVG_BUFF_LOOP

	return


; ReInitPic: Execute initalization code that is run at power-on or by a state refresh
;
; On entry: none
;
; On exit: PIC state udpated
;
ReInitPic
	; Initialize unique PIC control registers in page 0
	movlw		T1CON_INIT
	movwf		T1CON
	movlw		ADCON0_INIT
	movwf		ADCON0

	; Initialize unique PIC control registers in page 1
	bsf		STATUS,RP0	; Point to page 1 registers
	movlw		GPIO_OE
	movwf		TRISIO
	movlw		PIE1_SETUP
	movwf		PIE1
	movlw		WPU_INIT
	movwf		WPU
	movlw		IOC_INIT
	movwf		IOC
	movlw		ANSEL_INIT
	movwf		ANSEL
	bcf		STATUS,RP0	; Point back to page 0 registers

	return


; EvalAverage: Push the current ADC value (upper 8-bits) into the averaging buffer
; and compute a new 8-bit average value to use for the intensity.
;
; On entry: PushPtr points to next location in the averaging buffer to load data
;           ADRESH contains the high 8-bits of the ADC input
;
; On exit: PushPtr incremented
;          CurIntensity updated with new 8-bit average
;
EvalAverage
	; Push the new entry into current location
	movf		PushPtr,W
	movwf		FSR
	movf		ADRESH,W
	movwf		INDF
	call		IncPushPtr	; Increment pointer

	; Compute a new average to use as the intensity
	;
	; First sum all entries in the averaging buffer
	clrf		AvgHi
	clrf		AvgLo
	movlw		NUM_AVG_BUF_ENTRIES
	movwf		TempReg
	movlw		AvgBuf
	movwf		FSR
ADD_AVG_BUFF_LOOP
	movf		INDF,W
	addwf		AvgLo,F		; Add entry[n] to running lsb
	btfsc		STATUS,C
	incf		AvgHi,F		; Add carry to running msb
	incf		FSR,F
	decfsz		TempReg,F
	goto		ADD_AVG_BUFF_LOOP
	; Then shift the sum down to eight bits (in AvgLo)
	movlw		NUM_AVG_BUF_LOG
	movwf		TempReg
SHT_AVG_BUFF_LOOP
	bcf		STATUS,C
	rrf		AvgHi,F
	rrf		AvgLo,F
	decfsz		TempReg,F
	goto		SHT_AVG_BUFF_LOOP

	; Set the new intensity value
	movf		AvgLo,W
	movwf		CurIntensity

	return


; IncPushPtr: Increment the push pointer to the next location in the circular averaging
; buffer.
;
; On entry: PushPtr points to current location
;
; On exit: PushPtr points to next location
;
IncPushPtr
	incf		PushPtr,W
	andlw		AVG_BUF_PTR_MASK
	addlw		AvgBuf
	movwf		PushPtr
	return


	end
